"""Console script for acm_dl_hci_searcher."""
import sys

import click
from tabulate import tabulate
import textwrap
from acm_dl_searcher.__main__ import (_process_venue_data_from_doi,
                                      _get_collection_info,
                                      _get_entry_count,
                                      _search)
from acm_dl_searcher.search_operations import (GenericSearchFunction, GenericVenueFilter)
from acm_dl_searcher._utils import _display_results_html


@click.group()
def cli():
    """Console script for acm_dl_hci_searcher."""
    pass

@cli.command()
@click.argument("doi")
@click.option("--short-name", type=str, help="The short name to use for this venue", default=None)
@click.option("--force", type=bool, help="Force the short name if different short name is being provided.", default=False, is_flag=True)
def get(doi, short_name=None, force=False):
    """Get the information for a value. Expects a doi of a venue."""
    try:
        _process_venue_data_from_doi(doi, short_name, verify=True, force=force)
    except ValueError as e:
        print(e)


@cli.command()
@click.option("--full-path", type=bool, default=False, is_flag=True)
def list(full_path):
    """List all the details"""
    info, info_file = _get_collection_info()
    if full_path:
        print("The file location is: {} \n".format(info_file.parent))
        
    table = [[textwrap.fill(i["short_name"], 10), textwrap.fill(i["title"], 60), i["doi"], _get_entry_count(info_file.parent / name), info_file.parent / name if full_path else name] for name, i in info.items()]
    headers = ["Short Name", "Title", "DOI", "# of entries" ,"File"]
    print(tabulate(table, headers, tablefmt="fancy_grid"))


@cli.command()
@click.argument("pattern", type=str)
@click.option("--venue-short-name-filter", type=str, default=None)
@click.option("--print-abstract", type=bool, is_flag=True, default=False)
@click.option("--html", type=bool, is_flag=True, default=False, help="Show results on browser")
def search(pattern, venue_short_name_filter, print_abstract, html):
    """Search the database for matches"""
    results = _search(GenericSearchFunction(pattern, 0), GenericVenueFilter(venue_short_name_filter, None, None))
    formatted_results = [[result["doi"], result["year"], textwrap.fill(result["title"], 70), result["url"]] for result in results]
    if print_abstract:
        abstracts = [result["abstract"] for result in results]
        _formatted_results = []
        for i in range(len(formatted_results)):
            _formatted_results.append(formatted_results[i])
            _formatted_results.append(["", "", textwrap.fill(abstracts[i], 70), ""])
        formatted_results = _formatted_results
        header = ["DOI", "Year", "Title/Abstract", "URL"]
    else:
        header = ["DOI", "Year", "Title", "URL"]
    print(tabulate(formatted_results, header,tablefmt="fancy_grid"))
    if html:
        _display_results_html(pattern, results)


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



"""Main module."""
import requests
import re
from bs4 import BeautifulSoup
import bibtexparser
from pathlib import Path
import json
from tqdm import tqdm
import multiprocessing as mp
from tqdm.contrib.concurrent import process_map
from acm_dl_searcher._utils import _ensure_data_directory_exists, SEARCH_STRING

# TODO: better logging
# TODO: parallelize data collection
def _process_venue_data_from_doi(doi, short_name, overwrite=False, verify=False, force=False):
    """
    Takes a doi of a venue in acm and get the entries and abstract. The data will be cacheed in DATA_DIRECTORY. 
    Returns a dictionary containing the doi as keys and the details as values
    """    
    doi_file = _ensure_data_directory_exists() / (doi.replace("/", "_") + ".json")
    doi_entry = requests.get(f"http://doi.org/{doi}", headers={"Accept": "application/x-bibtex"}).text
    doi_title = bibtexparser.loads(doi_entry).entries[0]["title"]
    print(f"Title: {doi_title}")
    
    _update_collection_info(doi_file.name, doi, doi_title, short_name, force)

    doi_list_details = []
    if doi_file.exists():
        if not overwrite:
            with open(doi_file) as f:
                try:
                    doi_list_details = json.load(f)
                except json.decoder.JSONDecodeError:
                    pass
            if not verify:
                return doi_list_details
            print("File exists. Verifying if all entries exist.")
    
    doi = doi.replace("/", "%2F")
    search_string = SEARCH_STRING.format(key=doi, page_id=0)
    response = requests.get(search_string)

    # For debugging
    # with open("temp.html", "w") as f:
    #     f.write(response.text)

    doi_matcher = re.compile(r">https:\/\/doi\.org\/\d{2}\.\d{4}\/\d+\.\d+<")
    # A working curl for doi meta data: curl --location --silent --header "Accept: application/x-bibtex" https://doi.org/10.1145/3313831.3376868
    doi_list = [i.rstrip("<").lstrip(">") for i in doi_matcher.findall(response.text)]
    total_hits = int(BeautifulSoup(response.text, features="html.parser").find("span", {"class": "hitsLength"}).get_text())
    print(f"Found {total_hits} hits")

    max_workers = 2 * mp.cpu_count() + 2
    
    if total_hits > 50:
        page_id = 1
        search_strings = []
        while page_id * 50 < total_hits:
            search_strings.append(SEARCH_STRING.format(key=doi, page_id=page_id))
            page_id += 1

        print(f"Getting the urls for the {total_hits} entries")
        for urls in process_map(_get_doi_urls(doi_matcher), search_strings, max_workers=max_workers):
            doi_list.extend(urls)
        print()
        
    received_doi = [details["doi"] for details in doi_list_details]

    # Asynchronusly collect entries
    manager = mp.Manager()
    q = manager.Queue()    
    pool = mp.Pool(max_workers)

    #put listener to work first
    watcher = pool.apply_async(_bib_entry_collector, (q, doi_file, doi_list_details))

    print("Getting the entries")
    try:
        jobs = []
        for doi_url in doi_list:
            if doi_url.lstrip("https://doi.org/") in received_doi:
                continue
            jobs.append(pool.apply_async(_bib_entry_worker, (doi_url, q)))

        # collect results from the workers through the pool result queue
        for job in tqdm(jobs):
            job.get()

    finally:
        #now we are done, kill the listener
        q.put('kill')
        pool.close()
        pool.join()

        with open(doi_file) as f:
            doi_list_details = json.load(f)
            return doi_list_details


class _get_doi_urls:
    def __init__(self, doi_matcher):
        self.doi_matcher = doi_matcher

    def __call__(self, search_string):
        response = requests.get(search_string)
        return [i.rstrip("<").lstrip(">") for i in self.doi_matcher.findall(response.text)]


def _bib_entry_worker(doi_url, q):
    '''Get the doi entry and abstract of from a doi_url'''
    # Try five times before giving up
    for _ in range(5):
        try:
            details = bibtexparser.loads(requests.get(doi_url, headers={"Accept": "application/x-bibtex"}).text).entries[0]
            abstract_html = requests.get("https://dl.acm.org/doi/" + details["doi"]).text
            soup = BeautifulSoup(abstract_html, features="html.parser")
            abstract = soup.findAll("div", {"class": "abstractInFull"})[0].find("p").get_text()
            details["abstract"] = abstract
            q.put(details)
            break
        except KeyboardInterrupt:
            raise
        except Exception as e:
            print(e)


def _bib_entry_collector(q, doi_file, doi_list_details):
    '''listens for messages on the q, writes to file. '''
    with open(doi_file, 'w') as f:
        idx = 0
        while True:
            details = q.get()
            if details == 'kill' or idx % 50 == 0:
                with open(doi_file, "w") as f:
                    json.dump(doi_list_details, f, indent=4)
                if details == 'kill':
                    break
            doi_list_details.append(details)
            idx += 1
    

def _get_collection_info():
    """
    Reads the global information file and return the contents as adictionary and the path to the file itself.
    """
    info_file = _ensure_data_directory_exists() / "info.json"
    if info_file.exists():
        with open(info_file) as f:
            info = json.load(f)
    else:
        info = {}
    return info, info_file
        

def _update_collection_info(file_name, doi, title, short_name, force):
    """
    Adds an entry to the gloabl infomation file. If the entry exists, will be overwritten.
    """
    info, info_file = _get_collection_info()

    if file_name in info:
        if not force and short_name != info[file_name]["short_name"]:
            raise ValueError("Error: short-name provided does not match the one in database. " \
                             "If making sure all entries exist, run without the --short-name option to use the name in databse.\n"\
                             "To force the new name, use the option --force: acm-dl-searcher get <doi> --short-name <new-short-name> --force")
    else:
        if short_name is None:
            raise ValueError("Error: New entry: need paramter value for `short-name`. Pass value with option --short-name.")

    if short_name is not None and len(short_name) > 10:
        raise ValueError("Error: value for `--short-name` is too long (max 10 charachters).")
        
    info[file_name] = {"doi": doi, "title": title, "short_name": short_name}

    with open(info_file, "w") as f:
        json.dump(info, f, indent=4)


def _get_entry_count(doi_file):
    """
    Given a file path, will return the number of entries in that file. If the file reading files, returns None.
    """
    try:
        with open(doi_file) as f:
            content = json.load(f)
        return len(content)
    except:
        return None


def _search(search_fn, venue_filter=None):
    """
    Takes a two functions as parameters. And returns a set of entries from the complete databases that return true for both functions.
    
    :search_fn: A callable that takes one parameter and returns true or false. 
                The paramter passed to this callable will be the content of an entry.
    :vanue_filter: A callable that takes three paramters: (short_name, title, doi) and return a boolean.
    :return: Returns a list of entries.
    """
    info, info_file = _get_collection_info()

    if venue_filter is None:
        venue_filter = lambda short_name, title, doi:True

    entries = []
    root_dir = _ensure_data_directory_exists()
    print(f"Searching in `{root_dir}`")
    for doi_file, entry in info.items():
        if venue_filter(entry["short_name"], entry["title"], entry["doi"]):
            with open(root_dir / doi_file) as f:
                try:
                    full_content_list = json.load(f)
                except json.decoder.JSONDecodeError:
                    print(f"Warning: {entry['doi']} is empty.")
                    continue
            for content_dict in full_content_list:
                content = ":: ".join(content_dict.values())
                if search_fn(content):
                    entries.append(content_dict)

    return entries
                



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['acm_dl_searcher', 'tests']

package_data = \
{'': ['*'], 'acm_dl_searcher': ['templates/*']}

install_requires = \
['beautifulsoup4>=4.9.3,<5.0.0',
 'bibtexparser>=1.2.0,<2.0.0',
 'click>=7.1.2,<8.0.0',
 'fuzzysearch>=0.7.3,<0.8.0',
 'liquidpy>=0.6.2,<0.7.0',
 'minilog>=1.5,<2.0',
 'requests>=2.24.0,<3.0.0',
 'tabulate>=0.8.7,<0.9.0',
 'tqdm>=4.51.0,<5.0.0']

entry_points = \
{'console_scripts': ['acm-dl-searcher = acm_dl_searcher.cli:cli']}

setup_kwargs = {
    'name': 'acm-dl-searcher',
    'version': '0.1.0a2',
    'description': 'Top-level package for ACM DL Searcher.',
    'long_description': '===================\nACM DL Searcher\n===================\n\n\n.. image:: https://img.shields.io/pypi/v/acm_dl_hci_searcher.svg\n        :target: https://pypi.python.org/pypi/acm_dl_hci_searcher\n\n.. image:: https://readthedocs.org/projects/acm-dl-hci-searcher/badge/?version=latest\n        :target: https://acm-dl-hci-searcher.readthedocs.io/en/latest/?badge=latest\n        :alt: Documentation Status\n\nA simple command line tool to collect the entries on a particular venue and in acm and run searches on them.\n\n\nInstall\n-------\n\n.. code:: sh\n          \n   pip install acm-dl-searcher\n\nUsage\n--------\nIf getting the entries from the `CHI 16`_ Conference:\n\n* To get the entries of a particular venue:\n\n  .. code:: sh\n\n            acm-dl-searcher get 10.1145/2858036 --short-name "CHI 16""\n  \n  This will download all the entries and their abstracts. The short-name provided can be anything. The first parameter expected for ``amc-dl-searcher get`` is the doi of the venue.\n\n* To list all the venues saved:\n\n  .. code:: sh\n\n            acm-dl-searcher list\n\n* To search the from the saved venues:\n\n  .. code:: sh\n\n            acm-dl-searcher search "adaptive"\n\n  This will search all the venues obtained through ``acm-dl-searcher get``, and list out the paper and titles that contain the phrase "adaptive" in the abstract or title. Currently the searcher uses a fuzzy search with a maximum difference of 2.\n\n  To narrow the search to particular venue(s) use the option ``--venue-short-name-filter``:\n\n  .. code:: sh\n\n            acm-dl-searcher search "adaptive" --venue-short-name-filter "CHI"\n\n  This will list out the matches from venues whose `short name` contain "CHI".\n\n  To print out the abstracts as well use the option ``--print-abstracts``:\n  \n  .. code:: sh\n            \n     acm-dl-searcher search "adaptive" --print-abstracts\n\n  To view the results on the browser use the option ``--html``:\n  \n  .. code:: sh\n            \n     acm-dl-searcher search "adaptive" --html\n\n  \n\n\nCredits\n-------\n\nThis package was created with Cookiecutter_ and the `briggySmalls/cookiecutter-pypackage`_ project template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _`briggySmalls/cookiecutter-pypackage`: https://github.com/briggySmalls/cookiecutter-pypackage\n.. _`CHI 16`: https://dl.acm.org/doi/proceedings/10.1145/2858036\n',
    'author': 'Ahmed Shariff',
    'author_email': 'shariff.mfa@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ahmed-shariff/acm_dl_searcher',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from fuzzysearch import find_near_matches


class GenericSearchFunction:
    """A generic search function which returns if there is a fuzzy search match"""
    def __init__(self, pattern, max_l_dist=2):
        self.pattern = pattern
        self.max_l_dist = max_l_dist

    def __call__(self, content):
        return _generic_fuzzy_filter(content, self.pattern, self.max_l_dist)


class GenericVenueFilter:
    """ If any of short_name, title or doi matches a given input, the callable will return true"""
    def __init__(self, short_name_filter, title_filter, doi_filter):
        self.short_name_filter = short_name_filter
        self.title_filter = title_filter
        self.doi_filter = doi_filter
        
        if short_name_filter is None and title_filter is None and doi_filter is None:
            self._all_none = True
        else:
            self._all_none = False
            
    def  __call__(self, short_name, title, doi):
        if self._all_none:
            return True
        elif self.short_name_filter is not None and _generic_fuzzy_filter(short_name, self.short_name_filter, 0):
            return True
        elif self.title_filter is not None and _generic_fuzzy_filter(title, self.title_filter, 0):
            return True
        elif self.doi_filter is not None and _generic_fuzzy_filter(doi, self.doi_filter, 0):
            return True
        return False


def _generic_fuzzy_filter(string, pattern, max_l_dist=2):
    return len (find_near_matches(pattern, string, max_l_dist=max_l_dist)) > 0
    



import webbrowser
from liquid import Liquid
from pathlib import Path

DATA_DIRECTORY = Path.home() / ".acm_dl_data"
SEARCH_STRING = "https://dl.acm.org/action/doSearch?LimitedContentGroupKey={key}&pageSize=50&startPage={page_id}"


def _ensure_data_directory_exists(sub_dir=None):
    """Makes sure the data directory exists and returns the data directory path"""
    if sub_dir:
        path = DATA_DIRECTORY / sub_dir
    else:
        path = DATA_DIRECTORY

    if not path.exists():
        path.mkdir(parents=True)
    return path


def _display_results_html(pattern, search_results):
    with open(Path(__file__).parent / "templates/search_result.html") as f:
        ret = Liquid(f).render(tempName = f"Results for : {pattern} (found {len(search_results)})", items = search_results)

    out_file = _ensure_data_directory_exists("temp") / "search_results.html"
    with open(out_file, "w") as f:
        f.write(ret)

    webbrowser.open("file://" + str(out_file.absolute()))
        
